@@ -13,17 +13,21 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 import argparse
+import glob
 import itertools
 import logging
 import netrc
 import operator
 import os
+import shutil
 import subprocess
 import sys
 import tempfile
 from urllib import parse
+from uuid import uuid4
 
 import daiquiri
+
 import github
 
 
@@ -142,13 +146,18 @@ def git_get_title_and_message(begin, end):
     :return: number of commits, title, message
     """
     titles = git_get_log_titles(begin, end)
-
+    title = "Pull request for " + end
     if len(titles) == 1:
         title = titles[0]
-        message = git_get_commit_body(end)
+    pr_template = find_pull_request_template()
+    if pr_template:
+        message = get_pr_template_message(pr_template)
     else:
-        title = "Pull request for " + end
-        message = "\n".join(titles)
+        if len(titles) == 1:
+            message = git_get_commit_body(end)
+        else:
+            message = "\n".join(titles)
+
     return (len(titles), title, message)
 
 
@@ -158,7 +167,8 @@ def git_pull_request(target_remote=None, target_branch=None,
                      comment_on_update=True,
                      rebase=True,
                      force_editor=False,
-                     download=None):
+                     download=None,
+                     ignore_tag=False):
     branch = git_get_branch_name()
     if not branch:
         LOG.critical("Unable to find current branch")
@@ -226,7 +236,8 @@ def git_pull_request(target_remote=None, target_branch=None,
     else:
         fork_and_push_pull_request(g, repo, rebase, target_remote,
                                    target_branch, branch, user, title, message,
-                                   comment_on_update, comment, force_editor)
+                                   comment_on_update, comment,
+                                   force_editor, ignore_tag)
 
 
 def download_pull_request(g, repo, target_remote, pull_number):
@@ -245,9 +256,71 @@ def download_pull_request(g, repo, target_remote, pull_number):
         _run_shell_command(["git", "reset", "--hard", "FETCH_HEAD"])
 
 
+def edit_file_get_content_and_remove(filename):
+    editor = os.getenv("EDITOR")
+    if not editor:
+        LOG.warning(
+            "$EDITOR is unset, you will not be able to edit the "
+            "pull-request message")
+        editor = "cat"
+    status = os.system(editor + " " + filename)
+    if status != 0:
+        raise RuntimeError("Editor exited with status code %d" % status)
+    with open(filename, "r") as body:
+        content = body.read().strip()
+    os.unlink(filename)
+
+    return content
+
+
+def find_pull_request_template():
+    filename = "PULL_REQUEST_TEMPLATE*"
+    pr_template_paths = [
+        filename,
+        os.path.join(".github", filename),
+        os.path.join("docs", filename),
+    ]
+    for path in pr_template_paths:
+        templates = glob.glob(path)
+        if templates:
+            return templates[0]
+
+
+def get_pr_template_message(template):
+    fd, bodyfilename = tempfile.mkstemp()
+    shutil.copy(template, bodyfilename)
+    content = edit_file_get_content_and_remove(bodyfilename)
+
+    return content
+
+
+def edit_title_and_message(title, message):
+    fd, bodyfilename = tempfile.mkstemp()
+    with open(bodyfilename, "w") as body:
+        body.write(title + "\n\n")
+        body.write(message + "\n")
+    content = edit_file_get_content_and_remove(bodyfilename)
+
+    return parse_pr_message(content)
+
+
+def preserve_older_revision(branch, remote_to_push):
+    tag = "{}-{}".format(branch, uuid4())
+    _run_shell_command(["git", "tag", tag])
+    try:
+        _run_shell_command(["git", "push", remote_to_push, tag])
+    except RuntimeError:
+        LOG.error("Unable to push tag {} for previous revision".format(tag))
+    else:
+        LOG.info("Previous revision of patchset saved on tag {}".format(tag))
+
+    _run_shell_command(["git", "tag", "-d", tag])
+
+
 def fork_and_push_pull_request(g, repo_to_fork, rebase, target_remote,
                                target_branch, branch, user, title, message,
-                               comment_on_update, comment, force_editor):
+                               comment_on_update, comment,
+                               force_editor, ignore_tag):
 
     g_user = g.get_user()
 
@@ -277,13 +350,16 @@ def fork_and_push_pull_request(g, repo_to_fork, rebase, target_remote,
                  branch])
         except RuntimeError:
             LOG.error(
-                "It is likely that your change has a merge conflict. "
-                "You may resolve it in the working tree now as "
-                "described above and then run `git pull-request' again, or "
-                "if you do not want to resolve it yet (note that the "
-                "change can not merge until the conflict is resolved) "
-                "you may run `git rebase --abort' then `git pull-request -R' "
-                "to upload the change without rebasing.")
+                "It is likely that your change has a merge conflict.\n"
+                "You may resolve it in the working tree now as described "
+                "above.\n"
+                "Once done run `git pull-request' again.\n\n"
+                "If you want to abort conflict resolution, run "
+                "`git rebase --abort'.\n\n"
+                "Alternatively run `git pull-request -R' to upload the change "
+                "without rebase.\n"
+                "However the change won't able to merge until the conflict is "
+                "resolved.")
             return 37
 
     LOG.info("Force-pushing branch `%s' to remote `%s'",
@@ -296,6 +372,11 @@ def fork_and_push_pull_request(g, repo_to_fork, rebase, target_remote,
     if pulls:
         for pull in pulls:
             LOG.info("Pull-request updated:\n  %s", pull.html_url)
+
+            if force_editor:
+                title, message = edit_title_and_message(title or pull.title,
+                                                        message or pull.body)
+
             if title and message:
                 pull.edit(title=title, body=message)
                 LOG.debug("Updated pull-request title and message")
@@ -319,32 +400,16 @@ def fork_and_push_pull_request(g, repo_to_fork, rebase, target_remote,
         nb_of_commits, git_title, git_message = git_get_title_and_message(
             "%s/%s" % (target_remote, target_branch), branch)
 
+        title = title or git_title
+        message = message or git_message
+
         # Do not run an editor if there's only one commit or if both title and
         # message were specified
         if (force_editor or
            ((not title or not message) and nb_of_commits > 1) or
            (((title and not message) or
              (message and not title)) and nb_of_commits == 1)):
-            editor = os.getenv("EDITOR")
-            if not editor:
-                LOG.warning(
-                    "$EDITOR is unset, you will not be able to edit the "
-                    "pull-request message")
-                editor = "cat"
-
-            fd, bodyfilename = tempfile.mkstemp()
-            with open(bodyfilename, "w") as body:
-                body.write((title or git_title) + "\n\n")
-                body.write((message or git_message) + "\n")
-            os.system(editor + " " + bodyfilename)
-            with open(bodyfilename, "r") as body:
-                content = body.read().strip()
-            os.unlink(bodyfilename)
-
-            title, message = parse_pr_message(content)
-        else:
-            title = title or git_title
-            message = message or git_message
+            title, message = edit_title_and_message(title, message)
 
         if title is None:
             LOG.critical("Pull-request message is empty, aborting")
@@ -363,6 +428,9 @@ def fork_and_push_pull_request(g, repo_to_fork, rebase, target_remote,
         else:
             LOG.info("Pull-request created: " + pull.html_url)
 
+    if not ignore_tag:
+        preserve_older_revision(branch, remote_to_push)
+
 
 def _format_github_exception(action, exc):
         url = exc.data.get("documentation_url", "GitHub documentation")
@@ -414,6 +482,12 @@ def main():
         "--comment", "-C",
         help="Comment to publish when updating the pull-request"
     )
+    parser.add_argument(
+        "--no-tag-previous-revision",
+        action="store_true",
+        default=False,
+        help="Preserve older revision when pushing"
+    )
 
     args = parser.parse_args()
 
@@ -426,17 +500,23 @@ def main():
         level=logging.DEBUG if args.debug else logging.INFO,
     )
 
-    return git_pull_request(
-        target_remote=args.target_remote,
-        target_branch=args.target_branch,
-        title=args.title,
-        message=args.message,
-        comment_on_update=not args.no_comment_on_update,
-        comment=args.comment,
-        rebase=not args.no_rebase,
-        force_editor=args.force_editor,
-        download=args.download,
-    )
+    try:
+        return git_pull_request(
+            target_remote=args.target_remote,
+            target_branch=args.target_branch,
+            title=args.title,
+            message=args.message,
+            comment_on_update=not args.no_comment_on_update,
+            comment=args.comment,
+            rebase=not args.no_rebase,
+            force_editor=args.force_editor,
+            download=args.download,
+            ignore_tag=args.no_tag_previous_revision
+        )
+    except Exception as e:
+        if args.debug:
+            raise
+        LOG.error(e)
 
 
 if __name__ == '__main__':
